/* blctrl.c 
 *
 * Adapted from http://wiki.gp2x.org/wiki/GPIO_Programming_Example
 *
 * Copyright (C) 2005-2006, wiki.gp2x.org.
 *
 * Permission is granted to copy, distribute and/or modify this document
 * under the terms of the GNU Free Documentation License, Version 1.2
 * or any later version published by the Free Software Foundation;
 * with no Invariant Sections, no Front-Cover Texts, and no Back-Cover
 * Texts.  A copy of the license is included in the section entitled "GNU
 * Free Documentation License".
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stropts.h>
#include <errno.h>

#define GPIOHPINLVL	0x118E
#define GPIOHOUT	0x106E

static unsigned long  *mem32;
static unsigned short *mem16;
static int dev;

static int init(void)
{
  int prot = PROT_READ | PROT_WRITE;
  int flags = MAP_SHARED;

  dev = open("/dev/mem", O_RDWR);
  if (dev < 0) {
    perror("blctrl: init() open: ");
    return 0;
  }

  mem32 = mmap(0, 0x10000, prot, flags, dev, 0xc0000000);
  mem16 = (unsigned short *) mem32;

  if (mem32 == MAP_FAILED) {
    perror("blctrl: init() mmap: ");
    return 0;
  }

  return 1;
}

static void quit(void)
{
  if (close(dev) < 0) {
    perror("blctrl: quit() close: ");
  }
}

int main(void)
{
  if (!init()) {
    return 1;
  }

  printf("blctrl: lcd ");
  mem16[GPIOHOUT >> 1] ^= 4;
  printf("%s\n", mem16[GPIOHPINLVL >> 1] & 4 ? "off" : "on");

  quit();

  return 0;
}
