/* hotkeyd.c
 *
 * Copyright (C) 2006, bja <bja@kth.se>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/fb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

static int fb_size;
static unsigned char *fb_buf;
static unsigned char *fb_tmp;

static int fb_get_info(int fbdev, struct fb_var_screeninfo *info)
{
  int ret = 0;

  if (ioctl(fbdev, FBIOGET_VSCREENINFO, info) < 0) {
    perror("fb_get_info: ioctl");
    ret = -1;
  }

  return ret;
}

static int fb_open(const char *path)
{
  struct fb_var_screeninfo info;

  int fbdev = open(path, O_RDWR);
  if(fbdev < 0) {
    perror("fb_open: open");
    return -1;
  }

  if (fb_get_info(fbdev, &info) < 0) {
    close(fbdev);
    return -1;
  }
  
  int width = (int) info.xres;
  int height = (int) info.yres;
  int bpp = (int) info.bits_per_pixel;

  printf("%d %d %d\n", width, height, bpp);

  int mm_prot = PROT_READ | PROT_WRITE;
  int mm_flags = MAP_PRIVATE;

  fb_size = width * height * ((bpp + 7) >> 3);
  fb_buf = mmap(0, fb_size, mm_prot, mm_flags, fbdev, 0);

  if (fb_buf == MAP_FAILED) {
    perror("fb_open: mmap");
    close(fbdev);
    return -1;
  }

  return fbdev;
}

static int fb_close(int fbdev)
{
  int ret = 0;

  if (munmap(fb_buf, fb_size) < 0) {
    perror("fb_close: munmap");
    ret = -1;
  }

  if (close(fbdev) < 0) {
    perror("fb_close: close");
    ret = -1;
  }
  
  if (fb_tmp) {
    free(fb_tmp);
  }

  return ret;
}

static int fb_save(void)
{
  if (!fb_tmp) {
    fb_tmp = malloc(fb_size);
  }
  memcpy(fb_tmp, fb_buf, fb_size);
  return 0;
}

static int fb_load(void)
{
  memcpy(fb_buf, fb_tmp, fb_size);
  return 0;
}

static int fb_write(unsigned char *buf, int len, const char *filename)
{
  int ret = 0;

  int fd = open(filename, O_WRONLY | O_CREAT);
  if(fd < 0) {
    perror("fb_write: open");
    return -1;
  }
  
  if (write(fd, buf, len) < 0) {
    perror("fb_copy: write");
    ret = -1;
  }

  if (close(fd) < 0) {
    perror("fb_write: close");
    ret = -1;
  }

  return ret;
}


int main(int argc, char *argv[])
{
  if (argc < 2) {
    fprintf(stderr, "Usage: fbsave FILE\n");
    return -1;
  }

  int fbdev = fb_open("/dev/fb0");
  if (fbdev < 0) {
    return -1;
  }
  
  if (fb_write(fb_buf, fb_size, argv[1]) < 0) {
    return -1;
  }

  if (fb_close(fbdev) < 0) {
    return -1;
  }

  return 0;
}
