/* hotkeyd.c
 *
 * Copyright (C) 2006, bja <bja@kth.se>.
 * Copyright (C) 2005, rlyeh, GP2X Minimal Library.
 * copyrifht (c) 2006, kounch (hotkey personalization code)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <libgen.h>
#include <errno.h>

#define GP2X_UP		0x1
#define GP2X_LEFT	0x4
#define GP2X_DOWN	0x10
#define GP2X_RIGHT	0x40
#define GP2X_START	1<<8
#define GP2X_SELECT	1<<9
#define GP2X_L		1<<10
#define GP2X_R		1<<11
#define GP2X_A		1<<12
#define GP2X_B		1<<13
#define GP2X_X		1<<14
#define GP2X_Y		1<<15
#define GP2X_VOL_UP	1<<23
#define GP2X_VOL_DOWN	1<<22
#define GP2X_PUSH	1<<27

#define HOTKEY_TIMEOUT	1 /* number of seconds hotkey must be pressed */
#define HOTKEY_KEYS	(GP2X_L | GP2X_R)
#define HOTKEY_QUIT	(GP2X_START | GP2X_VOL_DOWN)

#ifdef _DEBUG
static int keycode[] = {
  GP2X_UP,     GP2X_LEFT,     GP2X_DOWN, GP2X_RIGHT,
  GP2X_START,  GP2X_SELECT,   GP2X_L,    GP2X_R,
  GP2X_A,      GP2X_B,        GP2X_X,    GP2X_Y,
  GP2X_VOL_UP, GP2X_VOL_DOWN, GP2X_PUSH
};
static char *keyname[] = {
  "GP2X_UP",     "GP2X_LEFT",     "GP2X_DOWN", "GP2X_RIGHT",
  "GP2X_START",  "GP2X_SELECT",   "GP2X_L",    "GP2X_R",
  "GP2X_A",      "GP2X_B",        "GP2X_X",    "GP2X_Y",
  "GP2X_VOL_UP", "GP2X_VOL_DOWN", "GP2X_PUSH"
};
#endif

static volatile unsigned short *mm_regs;
static volatile unsigned long  *mm_regl;
static int mm_dev;

static unsigned long gethkey(char *strHotkey);

static int gp2x_joy_init(void)
{
  void *mm_start = 0x00000;
  int mm_len = 0x10000;
  int mm_prot = PROT_READ | PROT_WRITE;
  int mm_flags = MAP_SHARED;
  int mm_offs = 0xC0000000;

  if (!mm_dev) {
    mm_dev = open("/dev/mem", O_RDWR);
    if (mm_dev < 0) {
      return -1;
    }
  }

  mm_regl = mmap(mm_start, mm_len, mm_prot, mm_flags, mm_dev, mm_offs);
  mm_regs = (unsigned short *) mm_regl;

  if (mm_regl == MAP_FAILED) {
    return -1;
  }

  return 0;
}

static int gp2x_joy_free(void)
{
  int ret = 0;

  if (munmap((void *) mm_regl, 0x10000) < 0) {
    ret = -1;
  }
  else {
    mm_regl = (void *) -1;
    mm_regs = (void *) -1;
  }

  if (close(mm_dev) < 0) {
    ret = -1;
  }
  else {
    mm_dev = -1;
  }

  return ret;
}

static unsigned long gp2x_joystick_read(void)
{
  short value = (mm_regs[0x1198>>1] & 0x00FF);
  switch (value) {
  case 0xFD: value=0xFA; break;
  case 0xF7: value=0xEB; break;
  case 0xDF: value=0xAF; break;
  case 0x7F: value=0xBE; break;
  }
 
  return ~((mm_regs[0x1184>>1] & 0xFF00) | value | (mm_regs[0x1186>>1] << 16));
}

int main(int argc, char *argv[])
{
  int hkey_timer = HOTKEY_TIMEOUT;
  unsigned long pad = 0;
  int status = 0;
  pid_t pid = 0;
  char *path = NULL;
  char *file = NULL;
  char *args = NULL;
  static unsigned long hotkeys, tempkey;

  /* We expect at least two arguments*/
  if (argc < 3) {
    fprintf(stderr, "Usage: hotkeyd HOTKEYS PROGRAM [ARGUMENTS]\n");
    exit(1);
  }

  printf("hotkeyd version 0.2\n");
  printf("Copyright (C) 2006, bja.\n");

  /*First argument is hotkey combination, separated by commas, use double quotes to set more than one key*/
  hotkeys=gethkey(strtok (argv[1], ","));
  
  while (tempkey=gethkey(strtok (NULL, ",")))
    hotkeys = hotkeys | tempkey;
  
  printf ("Hotkeys defined, code: %u\n",hotkeys);
  
  /*Second argument is the path to the file*/
  path = strdup(argv[2]);
  file = strdup(argv[2]);

  /*Third argument (if exists) is arguments to the file (use double quotes to put more than one argument)*/
  if (argc > 3) {
    args = argv[3];
  }

  gp2x_joy_init();

  while (1) {

    sleep(1);

    pad = gp2x_joystick_read();

#ifdef _DEBUG
    {int i; for (i=0; i<15; i++) {
      if (pad & keycode[i]) {
	printf("%s\n", keyname[i]);
      }
    }}
#endif

   /* if ((pad & HOTKEY_KEYS) == HOTKEY_KEYS && pid == 0) { */
    if ((pad & hotkeys) == hotkeys && pid == 0) {

      /* wait at least timer number of seconds before */

      if (hkey_timer--) {
	continue;
      }

      pid = fork();

      if (pid < 0) {
	perror("Error: call to fork() failed: ");
	break;
      } 
      else if (pid == 0) {

	/* launch the specified hotkey application */

	if (chdir(dirname(path)) < 0) {
	  perror("Error: call to chdir() failed: ");
	  break;
	}

	if (execl(file, file, args, NULL) < 0) {
	  perror("Error: call to execl() failed: ");
	  break;
	}
      }
      else {

	/* wait for hotkey application to terminate */

	if (waitpid(pid, &status, 0) < 0) {
	  perror("Error: call to waitpid() failed: ");
	  break;
	}

	pid = 0;
      }
    }
    else {

      /* reset timer if hotkey not pressed */

      hkey_timer = 1;
    }

    /* check for quit request */

    if ((pad & HOTKEY_QUIT) == HOTKEY_QUIT) {
	break;
    }
  }

  gp2x_joy_free();

  free(path);
  free(file);

  return 0;
}


/*Extract hotkey code from a string*/
static unsigned long gethkey(char *strHotkey)
{
  if (strHotkey == NULL)
    return 0;
  
  if (!strcmp(strHotkey,"UP"))
  {
    printf ("Detected UP\n");
    return (GP2X_UP);
  }
  if (!strcmp(strHotkey,"LEFT"))
  {
    printf ("Detected LEFT\n");
    return (GP2X_LEFT);
  }
  if (!strcmp(strHotkey,"DOWN"))
  {
    printf ("Detected DOWN\n");
    return (GP2X_DOWN);
  }
  if (!strcmp(strHotkey,"RIGHT"))
  {
    printf ("Detected RIGHT\n");
    return (GP2X_RIGHT);
  }
  if (!strcmp(strHotkey,"START"))
  {
    printf ("Detected START\n");
    return (GP2X_START);
  }
  if (!strcmp(strHotkey,"SELECT"))
  {
    printf ("Detected SELECT\n");
    return (GP2X_SELECT);
  }
  if (!strcmp(strHotkey,"L"))
  {
    printf ("Detected L\n");
    return (GP2X_L);
  }
  if (!strcmp(strHotkey,"R"))
  {
    printf ("Detected R\n");
    return (GP2X_R);
  }
  if (!strcmp(strHotkey,"A"))
  {
    printf ("Detected A\n");
    return (GP2X_A);
  }
  if (!strcmp(strHotkey,"B"))
  {
    printf ("Detected B\n");
    return (GP2X_B);
  }
  if (!strcmp(strHotkey,"X"))
  {
    printf ("Detected X\n");
    return (GP2X_X);
  }
  if (!strcmp(strHotkey,"Y"))
  {
    printf ("Detected Y\n");
    return (GP2X_Y);
  }
  if (!strcmp(strHotkey,"VOL_UP"))
  {
    printf ("Detected VOL_UP\n");
    return (GP2X_VOL_UP);
  }
   if (!strcmp(strHotkey,"VOL_DOWN"))
  {
    printf ("Detected VOL_DOWN\n");
    return (GP2X_VOL_DOWN);
  }
  if (!strcmp(strHotkey,"PUSH"))
  {
    printf ("Detected PUSH\n");
    return (GP2X_PUSH);
  }

  return (0);
};
