/*  EnergySaver for GP2X

	File: setbl.c
	
	Copyright (C) 2006  Kounch
	Parts Copyright (c) 2005 Rlyeh
	Parts Copyright (c) 2006 Hermes/PS2Reality

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>     //Entrada/Salida basica
#include <unistd.h>    //Para poder salir al menu de GP2X con execl, obtener pid
#include <stdlib.h>    //Para convertir de cadena a long, leer y escribir variables de entorno, etc.
#include <sys/mman.h>  //Para poder hacer mmap y acceder a los registros de memoria
#include <fcntl.h>     //Para poder tener acceso al dispositivo /dev/mem y /dev/fb
#include <string.h>    //Para concatenar cadenas

//Si compilamos para GP2X
#ifdef __GP2X__

#include "setbl.h" //Cabecera principal del programa
#include "cpulcd.h" //Parar ajustar la velocidad de reloj y el LCD

/* Definidas en cpculcd.c */
extern unsigned long gp2x_dev[8];
extern volatile unsigned short *gp2x_memregs;
extern volatile unsigned long  *gp2x_memregl;

/*Variables globales*/
char textoError[MAXLEN]; //Para almacenar la descripcion del ultimo error producido


/******************************************************************/
/*Rutina Main, aqui analizamos los argumentos de linea de comandos*/
/******************************************************************/

int main (int argc, char **argv) 
{
  char procesos[MAXLEN];
  char cad[MAXLEN];
  int horizontal,vertical;
  unsigned long tick;
  int i,corregir;
  unsigned velocidad;
  unsigned long tiempo; //El tiempo que hay que esperar (en segundos) entre iteraciones

  printf("\nSetBL "__VERSION_PROG__" para GP2X. Copyright 2006 Kounch\n");
  
  /*Comprobacion de argumentos*/
  if ( argc < 3 )
  {
    sprintf (textoError, "Error: setbl velocidad_cpu tiempo [excepcion1 excepcion2 ...]");
    muestraError();
    return(1); 
  }
  
  /*Se toma el primer argumento como dato de velocidad de CPU*/
  velocidad=atoi(argv[1]);
  if (velocidad > 166)
    velocidad=166;
  if (velocidad < 20)
    velocidad = 0;

  printf("Velocidad de CPU deseada: %i\n",velocidad);  

  /*Se toma el segundo argumento como dato de tiempo*/
  tiempo=atoi(argv[2]);
  if (tiempo < 0)
    tiempo=0;
  if (tiempo > 3600)
    tiempo=3600;

  printf("Tiempo de espera entre iteraciones: %i\n",tiempo);
  
  /* Se toma el argumento tercero y siguientes como excepciones a tener en cuenta*/
  if (argc>3)
  {
    sprintf(procesos," | grep -E '");
    for (i=3;i<argc;++i)
    { 
      sprintf(cad,"(%s)|",argv[i]);
      strcat(procesos,cad);
    }
    procesos[strlen(procesos)-1]=0;
    sprintf(cad,"' | grep -vE '(lsmod)|(grep)|(setbl)' > /dev/null");
    strcat(procesos,cad);
  }

  /*Inicializamos el acceso a los registros de memoria*/
  cpulcd_init();
  
  /* Esperamos el tiempo indicado, o ms, segn que se pulse o no algn control */
  do
  {
    if (tiempo)
    {  
      /* Bucle que se mantiene mientras se vaya detectando alguna entrada de los controles */
      for(i=0;i<(5*tiempo);i++)
      {   
        if (checkjoy())
          i=0;
        
        /* Esperar 200 milisegundos */
        /* Esto no carga la CPU de forma apreciable */
        usleep(200000);
        
        if (checkjoy())
          i=0;      
      }
    }
    else
    {
      /* Si el tiempo es 0, modo test con slo una iteracin */
      printf ("***TEST MODE***\nCheckJoy=%#00000000X\n",checkjoy());
    }
    printf ("***TEST MODE***\nCheckJoy=%#00000000X\n",checkjoy());


    /* En principio no vamos a cambiar nada */
    corregir=0;

    horizontal=gp2x_memregs[0x2906>>1];	
    vertical=gp2x_memregs[0x2908>>1];	
    printf ("Pantalla: %i x %i\n",horizontal,vertical);  

    /*Comprobando TV*/
    if(gp2x_memregs[0x2800>>1]&0x100)
    {
      printf ("TV-Out activado\n");
      
      if (gp2x_memregs[0x2818>>1]  == 287) //Comprobacion si la altura de la pantalla es para PAL
      {
        printf ("PAL detectado\n");
      }
      else if (gp2x_memregs[0x2818>>1]  == 239) //Comprobacion si la altura de la pantalla es para NTSC
      {
        printf ("NTSC detectado\n");
       }
    }
    else //Suponemos que esta en modo LCD (TV-Out desactivado)
    {
    printf ("TV-Out desactivado\n");
       
      //Como no estamos en salida de Tv, corregimos
      corregir=1;
    }
    
    /* Si hay excepciones que aplicar */
    if (argc>2)
    {
      /* Filtrado de procesos en ejecucin */
      sprintf (cad,"ps");
      strcat (cad,procesos);
      if (!system(cad))
      {
        corregir=0;
        printf("Se han detectado procesos.\n");
      } 

      /* Filtrado de mdulos del kernel cargados en memoria */
      sprintf (cad,"lsmod");
      strcat (cad,procesos);
        if (!system(cad))
        {
          corregir=0;
          printf("Se han detectado modulos de kernel.\n");
        } 
    }
    /* Si hemos determinado que hay que hacer ajustes, los aplicamos ahora */
    if (corregir)
    {
      printf ("Hay que hacer ajuste\n");
      bajoconsumo(velocidad);
    }
  }
  while (tiempo);
  
  cpulcd_deinit();
  printf ("Fin del programa\n");
  
  return(0);
}
#endif


/* Imprime un mensaje de error */
int muestraError(void)
  {
  int n,m;
  unsigned short col;
  
  fprintf (stderr, textoError);
  
  return (0);
}


/* Devuelve el valor de las entradas de los controles de la consola quitando eventos no deseados
   como  */
unsigned long checkjoy(void)
{
  unsigned long value, resultado;
  int i;

  value=(gp2x_memregs[0x1198>>1] & 0x00FF);

  if(value==0xFD) value=0xFA;
  if(value==0xF7) value=0xEB;
  if(value==0xDF) value=0xAF;
  if(value==0x7F) value=0xBE;
 
  resultado = ~((gp2x_memregs[0x1184>>1] & 0xFF00) | value | (gp2x_memregs[0x1186>>1] << 16));
 
  /* Filtramos contra las entradas de botones, ignorando cualquier otra cosa */
  resultado = resultado & 0x8C0FF55;
  
  return resultado;
}


/* Deja la consola en bajo consumo hasta que se pulse algn control */
int bajoconsumo(unsigned velocidad)
{
  unsigned cpu_speed, d_clock, estadoUCLK, estadoACLK;
  int volumen;
  
  /* Activamos el reposo*/
  printf ("Entrando en reposo\n");

  cpu_speed=get_speed_clock();
  d_clock=get_display_clock_div();
  volumen=gp2x_get_volume();

/*
  estadoUCLK=get_status_UCLK();
  estadoACLK=get_status_ACLK();
*/
  printf("Guardado estado original\n");

  /* Apagamos el LCD */
  set_display(0);
  printf("Apagado display\n");
  
/* Esto cuelga las consolas que no son First Edition 
  set_status_UCLK(0);// apaga los relojes
  set_status_ACLK(0);
  printf("Apagados relojes\n");
*/

  /* Si la velocidad es distinta de 0, ajustamos la CPU */
  if (velocidad)
  {
    gp2x_set_volume(0);
    printf ("Volumen bajado a 0\n");

    set_speed_clock(velocidad);
    printf ("CPU a a baja velocidad\n");
  
    set_DCLK_Div(3);
    set_920_Div(3);
    printf ("Cambiados los divisores de reloj\n");
  }

  printf ("En reposo\n");
  do
  {
    usleep(200000);
  }
  while (!checkjoy());

  printf ("Saliendo del reposo\n");

  /* Si la velocidad solicitada era distinta de 0, la restauramos */
  if (velocidad)
  {
    set_DCLK_Div(0);
    set_920_Div(0);
    printf ("Reajustados divisores de reloj\n");

    set_speed_clock(cpu_speed);
    printf ("Reajustada la CPU\n");

    gp2x_set_volume(volumen);
    printf ("Volumen subido de nuevo a %i\n",volumen);
  
/* Esto cuelga las consolas que no son First Edition 
    set_status_UCLK(estadoUCLK); // recupera los relojes
    set_status_ACLK(estadoACLK);
    printf("Reactivados relojes\n");
*/
    set_display_clock_div(d_clock);
    printf ("Recuperado el divisor de reloj de la LCD\n");
  }

  /* Reactivamos el LCD */
  set_display(1);
  printf ("Display On\n");

  printf ("Todo restablecido\n");
  return (0);
}
